function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

import ActionTypes from './actionTypes';

var id = function id(x) {
  return x;
};

/**
 * Promotes a value into a reducer.
 */
export function promote(reducer) {
  if (reducer instanceof Reducer) {
    return reducer;
  }
  if (typeof reducer === 'function') {
    var init = reducer(undefined, ActionTypes.INIT);
    return new Reducer(init, reducer);
  }
  return Reducer.of(reducer);
}

/**
  * Memoizes a presenter function's last input in order
  * to make sure new objects are not producer unnecessarily
  */
function memo1(f) {
  var lastInput = {};
  var lastOutput = void 0;
  return function memoizedPresenter(input) {
    if (input === lastInput) {
      return lastOutput;
    }
    lastInput = input;
    lastOutput = f(input);
    return lastOutput;
  };
}

/**
 * A reducer consists of an initial state, an update function
 * and a presenter function
 */

var Reducer = function () {
  function Reducer(init) {
    var update = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : id;
    var present = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : id;

    _classCallCheck(this, Reducer);

    this.init = init;
    this.update = update;
    this.present = memo1(present);
  }

  Reducer.prototype.map = function map(f) {
    var _this = this;

    return new Reducer(this.init, this.update, this.present === id ? f : function (x) {
      return f(_this.present(x));
    });
  };

  Reducer.prototype.contramap = function contramap(g) {
    var _this2 = this;

    return new Reducer(this.init, function (state, action) {
      return _this2.update(state, g(action));
    }, this.present);
  };

  Reducer.prototype.ap = function ap(r) {
    var _this3 = this;

    r = promote(r);
    return new Reducer({ x: this.init, f: r.init }, function (state, action) {
      var x = state.x,
          f = state.f;

      var next = {
        x: _this3.update(x, action),
        f: r.update(f, action)
      };
      return next.x === x && next.f === f ? state : next;
    }, function (_ref) {
      var x = _ref.x,
          f = _ref.f;
      return f(x);
    });
  };

  Reducer.prototype.promap = function promap(f, g) {
    return this.map(f).contramap(g);
  };

  Reducer.of = function of(x) {
    return new Reducer(x);
  };

  Reducer.prototype.extract = function extract() {
    return this.present(this.init);
  };

  Reducer.prototype.compose = function compose(r) {
    var _this4 = this;

    return new Reducer({ a: this.init, b: r.init }, function (state, action) {
      var a = state.a,
          b = state.b;

      var aNext = _this4.update(a, action);
      var bNext = r.update(b, _this4.present(aNext));
      return a === aNext && b === bNext ? state : { a: aNext, b: bNext };
    }, function (_ref2) {
      var b = _ref2.b;
      return r.present(b);
    });
  };

  Reducer.prototype.concat = function concat(r) {
    return this.compose(promote(r));
  };

  Reducer.prototype.extend = function extend(f) {
    var _this5 = this;

    return new Reducer(this.init, this.update, function (state) {
      return f(new Reducer(state, _this5.update, _this5.present));
    });
  };

  return Reducer;
}();

export default Reducer;


export var of = Reducer.of;

/** Lifts a function that runs on "regular" arguments
  * to work on reducers.
  */
export var lift = function lift(f) {
  return function () {
    for (var _len = arguments.length, reducers = Array(_len), _key = 0; _key < _len; _key++) {
      reducers[_key] = arguments[_key];
    }

    reducers = reducers.map(promote);
    return new Reducer(reducers.map(function (r) {
      return r.init;
    }), function (state, action) {
      var newstate = [];
      var hasChanged = false;
      for (var i = 0; i < reducers.length; i++) {
        newstate.push(reducers[i].update(state[i], action));
        hasChanged = hasChanged || newstate[i] !== state[i];
      }
      return hasChanged ? newstate : state;
    }, function (args) {
      var presentations = [];
      for (var i = 0; i < reducers.length; i++) {
        presentations.push(reducers[i].present(args[i]));
      }
      return f.apply(null, presentations);
    });
  };
};